use bevy::app::{App, CoreSet, Plugin};
use bevy::prelude::*;

/// An enum representing the current set of systems that should be running.
/// Changing states should be done by inserting an `iyes_loopless` `NextState`
/// resource with the value of one of these app states
#[derive(Clone, Copy, Eq, PartialEq, Ord, PartialOrd, Debug, Hash, States, Default)]
pub enum AppState {
	/// During the preload state, embedded resources will be registered with Bevy.
	/// Embedded resources will be attached to the asset system with the path
	/// `internal:{resource-name}`
	#[default]
	Preload,
	/// During the setup state, external resources will be registered by reading from
	/// the filesystem or the network. Progress can be displayed using the embedded
	/// resources loaded in the previous state
	Setup,
	/// The splash state exists solely to render attributions & logos before starting
	/// the game
	Splash,
	/// An initial landing page that will present players with options
	Menu,
	/// The in game state runs all of the actual gameplay logic. Most of the runtime
	/// will be spent here.
	InGame,
}

pub fn run_in_game(state: Res<State<AppState>>) -> bool {
	state.0 == AppState::InGame
}
pub fn run_in_menu(state: Res<State<AppState>>) -> bool {
	state.0 == AppState::Menu
}
pub fn run_in_splash(state: Res<State<AppState>>) -> bool {
	state.0 == AppState::Splash
}
pub fn run_in_setup(state: Res<State<AppState>>) -> bool {
	state.0 == AppState::Setup
}

#[derive(Clone, Copy, Eq, PartialEq, Ord, PartialOrd, Debug, Hash, SystemSet)]
#[system_set(base)]
pub enum FlowSet {
	// Any custom processing steps, such as "player actions", or "process AI"
	Cleanup,
}

#[derive(Clone, Copy, Eq, PartialEq, Ord, PartialOrd, Debug, Hash, SystemSet)]
#[system_set(base)]
pub enum FlushFlowCommands {
	// Create an associated flush step where actions will be applied, for each FlowSet
	FlushCleanup,
}

pub struct FlowSetPlugin;
impl Plugin for FlowSetPlugin {
	fn build(&self, app: &mut App) {
		app.configure_sets(
			(
				CoreSet::UpdateFlush,
				FlowSet::Cleanup,
				FlushFlowCommands::FlushCleanup,
				CoreSet::PostUpdate,
			)
				.chain(),
		)
		.add_system(apply_system_buffers.in_base_set(FlushFlowCommands::FlushCleanup));
	}
}

pub trait AppScheduleExt {
	fn add_enter_system<M>(
		&mut self,
		state: impl States,
		system: impl IntoSystemAppConfig<M>,
	) -> &mut Self;
	fn add_enter_systems<M>(
		&mut self,
		state: impl States,
		system: impl IntoSystemAppConfigs<M>,
	) -> &mut Self;
	fn add_exit_system<M>(
		&mut self,
		state: impl States,
		system: impl IntoSystemAppConfig<M>,
	) -> &mut Self;
	fn add_exit_systems<M>(
		&mut self,
		state: impl States,
		system: impl IntoSystemAppConfigs<M>,
	) -> &mut Self;
}

impl AppScheduleExt for App {
	fn add_enter_system<M>(
		&mut self,
		state: impl States,
		system: impl IntoSystemAppConfig<M>,
	) -> &mut Self {
		self.add_system(system.in_schedule(OnEnter(state)))
	}

	fn add_enter_systems<M>(
		&mut self,
		state: impl States,
		systems: impl IntoSystemAppConfigs<M>,
	) -> &mut Self {
		self.add_systems(systems.in_schedule(OnEnter(state)))
	}

	fn add_exit_system<M>(
		&mut self,
		state: impl States,
		system: impl IntoSystemAppConfig<M>,
	) -> &mut Self {
		self.add_system(system.in_schedule(OnExit(state)))
	}

	fn add_exit_systems<M>(
		&mut self,
		state: impl States,
		systems: impl IntoSystemAppConfigs<M>,
	) -> &mut Self {
		self.add_systems(systems.in_schedule(OnExit(state)))
	}
}
